/* This file is part of Mailfromd.
   Copyright (C) 2005-2011, 2015 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdlib.h>
#include <errno.h>
#include "libmf.h"
#include "mailutils/alloc.h"


struct cache_list_elt {
	struct cache_list_elt *next;
	struct mf_option_cache *opt;
	size_t size;
};

static struct cache_list_elt *head, *tail;

static struct mf_option_cache *
optcache_dup(struct mf_option_cache *tab, size_t size)
{
	struct mf_option_cache *newtab = mu_calloc(size, sizeof(*tab));
	size_t i;

	for (i = 0; i < size; i++) {
		newtab[i] = tab[i];
		newtab[i].name = mu_strdup(newtab[i].name);
		newtab[i].isset = 0;
	}
	return newtab;
}

void
mf_optcache_add(struct mf_option_cache *tab, size_t size, int flags)
{
	struct cache_list_elt *elt;

	elt = mu_alloc(sizeof(*elt));
	elt->next = NULL;
	if (tail)
		tail->next = elt;
	else
		head = elt;
	tail = elt;

	/* If the array is null-terminated, compute its size */
	if (flags & MF_OCF_NULL)
		for (size = 0; tab[size].name; size++)
			;

	/* Unless it is static, make a copy */
	tab = optcache_dup(tab, size);

	elt->opt = tab;
	elt->size = size;
}

static struct mf_option_cache *
find_option(char const *name)
{
	struct cache_list_elt *elt;

	for (elt = head; elt; elt = elt->next) {
		size_t i;
		struct mf_option_cache *opt;
		
		for (i = 0, opt = elt->opt; i < elt->size; i++, opt++) {
			if (strcmp(opt->name, name) == 0)
				return opt;
		}
	}
	return NULL;
}

int
mf_optcache_set_option(char const *name, char const *value)
{
	int rc;
	struct mf_option_cache *p = find_option(name);
	if (!p) {
		errno = ENOENT;
		return 1;
	}
	rc = p->handler(name, &p->value, value);
	if (rc == 0)
		p->isset = 1;
	return rc;
}

void
mf_optcache_flush()
{
	struct cache_list_elt *elt;

	for (elt = head; elt; elt = elt->next) {
		size_t i;
		struct mf_option_cache *opt;
		
		for (i = 0, opt = elt->opt; i < elt->size; i++, opt++)
			if (opt->isset && opt->set)
				opt->set(&opt->value);
	}
}


int
mf_option_string(char const *opt, union mf_option_value *val, char const *arg)
{
	free(val->ov_string);
	val->ov_string = strdup(arg);
	return 0;
}

int
mf_option_boolean(char const *opt, union mf_option_value *val, char const *arg)
{
	int b;

	if (strcmp (arg, "yes") == 0
	    || strcmp (arg, "true") == 0
	    || strcmp (arg, "t") == 0)
		b = 1;
	else if (strcmp (arg, "no") == 0
		 || strcmp (arg, "false") == 0
		 || strcmp (arg, "nil") == 0)
		b = 0;
	else {
		char *p;
		
		b = strtoul(arg, &p, 10);
		if (*p) {
			errno = EINVAL;
			return 1;
		}
	}
	
	val->ov_bool = b;
	return 0;
}

int
mf_option_timeout(char const *opt, union mf_option_value *val, char const *arg)
{
	time_t interval;
	const char *endp;
	
	if (parse_time_interval(arg, &interval, &endp)) {
		mu_error(_("%s: unrecognized time format (near `%s')"),
			 opt, endp);
		return 1;
	}
	val->ov_time = interval;
	return 0;
}

int
mf_option_size(char const *opt, union mf_option_value *val, char const *arg)
{
	char *p;
	unsigned long n = strtoul(arg, &p, 10);
	if (*p)
		return 1;
	val->ov_size = n;
	return 0;
}
	
